// Copyright © 2023 ab180. All rights reserved.

#include "AirbridgeIOSCore.h"
#include "UObject/UObjectIterator.h"

#include "IOSAppDelegate.h"

#import "Platforms/iOS/Native/AUAppDelegate.h"
#import "Platforms/iOS/Native/AUAppSetting.h"
#import "Platforms/iOS/Native/Utils/AUHex.h"

#import <UIKit/UIKit.h>
#import <Airbridge/Airbridge.h>

#include "Blueprints/AirbridgeCallbacks.h"

FString AirbridgeIOSCore::DeeplinkData;
TMap<FString, FString> AirbridgeIOSCore::AttributionData;

AirbridgeIOSCore::AirbridgeIOSCore()
{
    [AUAppDelegate load];
}

AirbridgeIOSCore::~AirbridgeIOSCore()
{
    [AUAppDelegate unregisterLifeCycleListener];
}

#pragma region Core

void AirbridgeIOSCore::EnableSDK()
{
	[Airbridge enableSDK];
}

void AirbridgeIOSCore::DisableSDK()
{
	[Airbridge disableSDK];
}

bool AirbridgeIOSCore::IsSDKEnabled()
{
	return Airbridge.isSDKEnabled;
}

#pragma endregion

#pragma region Privacy

void AirbridgeIOSCore::StartTracking()
{
	[Airbridge startTracking];
}

void AirbridgeIOSCore::StopTracking()
{
    [Airbridge stopTracking];
}

bool AirbridgeIOSCore::IsTrackingEnabled()
{
    return Airbridge.isTrackingEnabled;
}

#pragma endregion

#pragma region Deeplink

void AirbridgeIOSCore::SetOnDeeplinkReceived(const AirbridgeStringParamFunction OnDeeplinkReceived)
{
	if (HandleDeeplinkListener) { return; }

    HandleDeeplinkListener = OnDeeplinkReceived;
	
	if (DeeplinkData.IsEmpty())
	{
		[Airbridge handleDeferredDeeplinkOnSuccess:^(NSURL* url) {
			if (url != nil) {
				NSString *urlString = [url absoluteString];
				ProcessDeeplink(FString(urlString));
			}
		}];
	}
	else
	{
		HandleDeeplink();
	}
}

#pragma endregion

#pragma region Data Collection

void AirbridgeIOSCore::SetUserID(const FString& Id)
{
	[Airbridge setUserID:Id.GetNSString()];
}

void AirbridgeIOSCore::ClearUserID()
{
	[Airbridge clearUserID];
}

void AirbridgeIOSCore::SetUserEmail(const FString& Email)
{
	[Airbridge setUserEmail:Email.GetNSString()];
}

void AirbridgeIOSCore::ClearUserEmail()
{
	[Airbridge clearUserEmail];
}

void AirbridgeIOSCore::SetUserPhone(const FString& Phone)
{
	[Airbridge setUserPhone:Phone.GetNSString()];
}

void AirbridgeIOSCore::ClearUserPhone()
{
	[Airbridge clearUserPhone];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const int32 Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const int64 Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const float Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const double Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const bool Value)
{
    [Airbridge setUserAttributeWithKey:Key.GetNSString() value:@(Value)];
}

void AirbridgeIOSCore::SetUserAttribute(const FString& Key, const FString& Value)
{
	[Airbridge setUserAttributeWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveUserAttribute(const FString& Key)
{
    [Airbridge removeUserAttributeWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearUserAttributes()
{
	[Airbridge clearUserAttributes];
}

void AirbridgeIOSCore::SetUserAlias(const FString& Key, const FString& Value)
{
	[Airbridge setUserAliasWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveUserAlias(const FString& Key)
{
	[Airbridge removeUserAliasWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearUserAlias()
{
	[Airbridge clearUserAlias];
}

void AirbridgeIOSCore::ClearUser()
{
	[Airbridge clearUser];
}

void AirbridgeIOSCore::SetDeviceAlias(const FString& Key, const FString& Value)
{
	[Airbridge setDeviceAliasWithKey:Key.GetNSString() value:Value.GetNSString()];
}

void AirbridgeIOSCore::RemoveDeviceAlias(const FString& Key)
{
    [Airbridge removeDeviceAliasWithKey:Key.GetNSString()];
}

void AirbridgeIOSCore::ClearDeviceAlias()
{
	[Airbridge clearDeviceAlias];
}

void AirbridgeIOSCore::RegisterPushToken(const FString& Token)
{
	NSString *tokenString = Token.GetNSString();
	NSData *tokenData = [AUHex dataFromHexString:tokenString];
	[Airbridge registerPushToken:tokenData];
}

#pragma endregion

#pragma region Event

void AirbridgeIOSCore::TrackEvent(const FString& Category, const UAirbridgeMap* SemanticAttributes, const UAirbridgeMap* CustomAttributes)
{
	[Airbridge
		trackEventWithCategory:Category.GetNSString()
		semanticAttributes:(SemanticAttributes != nullptr) ? SemanticAttributes->ConvertToNSDictionary() : nil
		customAttributes: (CustomAttributes != nullptr) ? CustomAttributes->ConvertToNSDictionary() : nil
	];
}

#pragma endregion

#pragma region Placement

void AirbridgeIOSCore::Click(const FString& TrackingLink, const AirbridgeVoidParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	NSURL *url = [NSURL URLWithString:TrackingLink.GetNSString()];
	
	[Airbridge
		clickWithTrackingLink:url
		onSuccess:^{
			if (OnSuccess) { OnSuccess(); }
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) { OnFailure(error.localizedDescription); }
		}];
}

void AirbridgeIOSCore::Impression(const FString& TrackingLink, const AirbridgeVoidParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	NSURL *url = [NSURL URLWithString:TrackingLink.GetNSString()];
    
	[Airbridge
		impressionWithTrackingLink:url
		onSuccess:^{
			if (OnSuccess) { OnSuccess(); }
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) { OnFailure(error.localizedDescription); }
		}];
}

void AirbridgeIOSCore::CreateTrackingLink(
	const FString& Channel,
	const UAirbridgeMap* Option,
	const AirbridgeTrackingLinkParamFunction OnSuccess,
	const AirbridgeStringParamFunction OnFailure
)
{
	[Airbridge
		createTrackingLinkWithChannel:Channel.GetNSString()
		option:(Option != nullptr) ? Option->ConvertToNSDictionary() : nil
		onSuccess:^(AirbridgeTrackingLink * _Nonnull trackingLink) {
			FAirbridgeTrackingLink result;
			result.ShortURL = FString([trackingLink.shortURL.absoluteString UTF8String]);
			result.QrcodeURL = FString([trackingLink.qrcodeURL.absoluteString UTF8String]);

			OnSuccess(result);	
		}
		onFailure:^(NSError * _Nonnull error) {
			OnFailure(FString(UTF8_TO_TCHAR([error.localizedDescription UTF8String])));
		}
    ];
}

#pragma endregion

#pragma region Publication

void AirbridgeIOSCore::FetchDeviceUUID(const AirbridgeStringParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	[Airbridge
		fetchDeviceUUIDOnSuccess:^(NSString * _Nonnull deviceUUID) {
			if (OnSuccess) {
				const char* cString = [deviceUUID UTF8String];
				OnSuccess(FString(UTF8_TO_TCHAR(cString)));
			}
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) {
				const char* cString = [error.localizedDescription UTF8String];
				OnFailure(FString(UTF8_TO_TCHAR(cString)));
			}
		}];
}

void AirbridgeIOSCore::FetchAirbridgeGeneratedUUID(const AirbridgeStringParamFunction OnSuccess, const AirbridgeStringParamFunction OnFailure)
{
	[Airbridge
		fetchAirbridgeGeneratedUUIDOnSuccess:^(NSString * _Nonnull airbridgeGeneratedUUID) {
			if (OnSuccess) {
				const char* cString = [airbridgeGeneratedUUID UTF8String];
				OnSuccess(FString(UTF8_TO_TCHAR(cString)));
			}
		}
		onFailure:^(NSError * _Nonnull error) {
			if (OnFailure) {
				const char* cString = [error.localizedDescription UTF8String];
				OnFailure(FString(UTF8_TO_TCHAR(cString)));
			}
		}];
}

void AirbridgeIOSCore::SetOnAttributionReceived(const AirbridgeMapParamFunction OnAttributionReceived)
{
	if (AirbridgeOnAttributionResultReceiveListener) return;

    AirbridgeOnAttributionResultReceiveListener = OnAttributionReceived;

	HandleOnAttributionReceived();
}

#pragma endregion

#pragma region Hybrid

FString AirbridgeIOSCore::CreateWebInterfaceScript(const FString& WebToken, const FString& PostMessageScript)
{
    NSString *script = [Airbridge
        createWebInterfaceScriptWithWebToken:WebToken.GetNSString()
        postMessageScript:PostMessageScript.GetNSString()
    ];
    
    const char* cString = [script UTF8String];
    return FString(cString);
}

void AirbridgeIOSCore::HandleWebInterfaceCommand(const FString& Command)
{
    [Airbridge handleWebInterfaceCommand:Command.GetNSString()];
}

#pragma endregion

void AirbridgeIOSCore::ProcessDeeplink(const FString& Uri)
{
	DeeplinkData = Uri;
	HandleDeeplink();
}

void AirbridgeIOSCore::HandleDeeplink()
{
	if (DeeplinkData.IsEmpty()) return;

	// Use Blueprints
	bool UseBroadcast = false;
	for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
	{
		Itr->OnDeeplinkReceived.Broadcast(DeeplinkData);
		UseBroadcast = true;
	}
	if (UseBroadcast)
	{
		DeeplinkData.Reset();
	}

	// Use C++
	if (HandleDeeplinkListener)
	{
		HandleDeeplinkListener(DeeplinkData);
		DeeplinkData.Reset();
	}
}

void AirbridgeIOSCore::ProcessOnAttributionResultReceived(const TMap<FString, FString>& Data)
{
	AttributionData = Data;
	HandleOnAttributionReceived();
}

void AirbridgeIOSCore::HandleOnAttributionReceived()
{
	if (AttributionData.Num() == 0) return;

	// Use Blueprints
	bool UseBroadcast = false;
	for (TObjectIterator<UAirbridgeCallbacks> Itr; Itr; ++Itr)
	{
		Itr->OnAttributionReceived.Broadcast(FAirbridgeBlueprintMap { AttributionData });
		UseBroadcast = true;
	}
	if (UseBroadcast)
	{
		AttributionData.Empty();
	}

	// Use C++
	if (AirbridgeOnAttributionResultReceiveListener)
	{
		AirbridgeOnAttributionResultReceiveListener(AttributionData);
		AttributionData.Empty();
	}
}

//MARK: - IOSAppDelegate+AirbridgeAppDelegate
@interface IOSAppDelegate (AirbridgeAppDelegate)

@end

@implementation IOSAppDelegate (AirbridgeAppDelegate)

+ (void)processDeeplinkURL:(NSURL * __nullable)url {
    if (!url) { return; }
    
    [self postDeeplinkCallback:url.absoluteString];
}

+ (void)postDeeplinkCallback:(NSString *)deeplinkString {
    FString string(deeplinkString);
    
    AsyncTask(ENamedThreads::GameThread, [string = MoveTemp(string)]() mutable
    {
        if (IOS_CORE_POINTER) { IOS_CORE_POINTER->ProcessDeeplink(string); }
    });
}


- (BOOL)application:(UIApplication *)application openURL:(NSURL *)url options:(NSDictionary<UIApplicationOpenURLOptionsKey, id>*)options
{
	[Airbridge trackDeeplinkWithUrl:url];
    
    BOOL isHandled = [Airbridge handleDeeplinkWithUrl:url onSuccess:^(NSURL* url) {
        if (!GIsRunning) {
            AirbridgeIOSCore::DeeplinkData = FString(url.absoluteString);
        }
        else {
            [IOSAppDelegate processDeeplinkURL:url];
        }
	}];
    
    if (isHandled || AUAppSetting.isHandleAirbridgeDeeplinkOnly) { return YES; }
    if (!GIsRunning) {
        AirbridgeIOSCore::DeeplinkData = FString(url.absoluteString);
    } else {
        [IOSAppDelegate processDeeplinkURL:url];
    }
    
	return YES;
}

- (BOOL)application:(UIApplication*)application continueUserActivity:(NSUserActivity*)userActivity restorationHandler:(void (^)(NSArray* _Nullable))restorationHandler
{
	[Airbridge trackDeeplinkWithUserActivity:userActivity];
    NSURL *url = userActivity.webpageURL;
    if (!url) { return YES; }
    
    BOOL isHandled = [Airbridge handleDeeplinkWithUserActivity:userActivity onSuccess:^(NSURL* url) {
        if (!GIsRunning) {
            AirbridgeIOSCore::DeeplinkData = FString(url.absoluteString);
        }
        else {
            [IOSAppDelegate processDeeplinkURL:url];
        }
	}];
    
    if (isHandled || AUAppSetting.isHandleAirbridgeDeeplinkOnly) { return YES; }
    if (!GIsRunning) {
        AirbridgeIOSCore::DeeplinkData = FString(url.absoluteString);
    } else {
        [IOSAppDelegate processDeeplinkURL:url];
    }
	
	return YES;
}

@end
